@@ -1,22 +1,42 @@
-from datetime import datetime, date
+from datetime import datetime, date, timezone
+import inspect
+from itertools import islice
 import logging
 from pathlib import Path
 from random import Random
-from subprocess import check_call
+from subprocess import check_call, run, PIPE
 import string
 import sys
 import time
 import timeit
 from typing import NamedTuple, Iterator, Optional, List, Set, Tuple, cast, Iterable, Dict, Any, Union
 
+from more_itertools import one
+
 import pytz
 import pytest  # type: ignore
 
-from .. import cachew, get_logger, PRIMITIVES, NTBinder, CachewException, Types, Values
+from .. import cachew, get_logger, PRIMITIVES, NTBinder, CachewException, Types, Values, settings
 
 
 logger = get_logger()
 
+@pytest.fixture(autouse=True)
+def throw_on_errors():
+    # a more reasonable default for tests
+    settings.THROW_ON_ERROR = True
+    yield
+
+
+@pytest.fixture
+def restore_settings():
+    orig = {k: v for k, v in settings.__dict__.items() if not k.startswith('__')}
+    try:
+        yield
+    finally:
+        for k, v in orig.items():
+            setattr(settings, k, v)
+
 
 @pytest.fixture
 def tdir(tmp_path):
@@ -84,7 +104,7 @@ def test_custom_hash(tdir):
 
     @cachew(
         cache_path=tdir,
-        hashf=lambda path: path.stat().st_mtime  # when path is update, underlying cache would be discarded
+        depends_on=lambda path: path.stat().st_mtime  # when path is update, underlying cache would be discarded
     )
     def data(path: Path) -> Iterable[UUU]:
         nonlocal calls
@@ -140,6 +160,56 @@ def inner(_it, _timer{init}):
     assert t < 2.0, 'should be pretty much instantaneous'
 
 
+def test_cache_path(tdir):
+    calls = 0
+    def orig() -> Iterable[int]:
+        nonlocal calls
+        yield 1
+        yield 2
+        calls += 1
+
+    fun = cachew(tdir / 'non_existent_dir' / 'cache_dir')(orig)
+    assert list(fun()) == [1, 2]
+    assert calls == 1
+    assert list(fun()) == [1, 2]
+    assert calls == 1
+
+    # dir by default
+    cdir = tdir / 'non_existent_dir' / 'cache_dir'
+    assert cdir.is_dir()
+    cfile = one(cdir.glob('*'))
+    assert cfile.name.startswith('cachew.tests.test_cachew:test_cache_path.')
+
+    # treat None as "don't cache"
+    fun = cachew(cache_path=None)(orig)
+    assert list(fun()) == [1, 2]
+    assert calls == 2
+    assert list(fun()) == [1, 2]
+    assert calls == 3
+
+    f = tdir / 'a_file'
+    f.touch()
+    fun = cachew(cache_path=f)(orig)
+    assert list(fun()) == [1, 2]
+    assert calls == 4
+    assert list(fun()) == [1, 2]
+    assert calls == 4
+
+    fun = cachew(tdir / 'name', force_file=True)(orig)
+    assert list(fun()) == [1, 2]
+    assert calls == 5
+    assert list(fun()) == [1, 2]
+    assert calls == 5
+
+    # if passed force_file, also treat as file
+    assert (tdir / 'name').is_file()
+
+    # TODO this won't work at the moment
+    # f.write_text('garbage')
+    # not sure... on the one hand could just delete the garbage file and overwrite with db
+    # on the other hand, wouldn't want to delete some user file by accident
+
+
 class TE2(NamedTuple):
     value: int
     uuu: UUU
@@ -330,7 +400,6 @@ def test_transaction(tdir):
     assert list(get_data(1)) == exp
 
     # TODO test that hash is unchanged?
-    import pytest # type: ignore
     with pytest.raises(TestError):
         list(get_data(2))
 
@@ -443,9 +512,49 @@ def test_dataclass(tdir):
     assert list(get_dataclasses()) == [Test(field=i) for i in range(5)]
 
 
-def test_types(tdir):
+def test_dates(tmp_path):
     from dataclasses import dataclass
     @dataclass
+    class X:
+        d1: datetime
+        d2: datetime
+        d3: datetime
+        d4: datetime
+        d5: datetime
+
+    tz = pytz.timezone('Europe/London')
+    dwinter = datetime.strptime('20200203 01:02:03', '%Y%m%d %H:%M:%S')
+    dsummer = datetime.strptime('20200803 01:02:03', '%Y%m%d %H:%M:%S')
+
+    x = X(
+        d1=tz.localize(dwinter),
+        d2=tz.localize(dsummer),
+        d3=dwinter,
+        d4=dsummer,
+        d5=dsummer.replace(tzinfo=timezone.utc),
+    )
+
+    @cachew(tmp_path)
+    def fun() -> Iterable[X]:
+        yield x
+
+    assert one(fun()) == x
+    assert one(fun()) == x
+
+    # make sure the actuall tzinfo is preserved... otherwise we might end up with raw offsets and lose some info
+    r = one(fun())
+    # attempting to preserve pytz zone names is a bit arbitrary
+    # but on the other hand, they will be in python 3.9, so I guess it's ok
+    assert r.d1.tzinfo.zone == x.d1.tzinfo.zone # type: ignore
+    assert r.d2.tzinfo.zone == x.d2.tzinfo.zone # type: ignore
+    assert r.d3.tzname() is None
+    assert r.d4.tzname() is None
+    assert r.d5.tzinfo is timezone.utc
+
+
+def test_types(tdir):
+    from dataclasses import dataclass, asdict
+    @dataclass
     class Test:
         an_int : int
         a_bool : bool
@@ -454,9 +563,8 @@ def test_types(tdir):
         a_dt   : datetime
         a_date : date
         a_json : Dict[str, Any]
+        an_exc : Exception
 
-        # temporary commented: experimental feature
-        # a_exc  : Optional[Exception] # TODO ugh. exceptions can't be compared really...
     # pylint: disable=no-member
     assert len(Test.__annotations__) == len(PRIMITIVES) # precondition so we don't forget to update test
 
@@ -469,14 +577,21 @@ def test_types(tdir):
         a_dt   =datetime.now(tz=tz),
         a_date =datetime.now().replace(year=2000).date(),
         a_json ={'a': True, 'x': {'whatever': 3.14}},
+        an_exc =RuntimeError('error!', 123),
     )
 
     @cachew(tdir)
     def get() -> Iterator[Test]:
         yield obj
 
-    assert list(get()) == [obj]
-    assert list(get()) == [obj]
+    def H(t: Test):
+        # Exceptions can't be directly compared.. so this kinda helps
+        d = asdict(t)
+        d['an_exc'] = d['an_exc'].args
+        return d
+
+    assert H(one(get())) == H(obj)
+    assert H(one(get())) == H(obj)
 
 
 # TODO if I do perf tests, look at this https://docs.sqlalchemy.org/en/13/_modules/examples/performance/large_resultsets.html
@@ -495,7 +610,7 @@ def test_primitive(tmp_path: Path):
 class O(NamedTuple):
     x: int
 
-def test_default(tmp_path: Path):
+def test_default_arguments(tmp_path: Path):
     class HackHash:
         def __init__(self, x: int) -> None:
             self.x = x
@@ -505,19 +620,48 @@ def test_default(tmp_path: Path):
 
     hh = HackHash(1)
 
-    @cachew(tmp_path, hashf=lambda param: param.x)
-    def fun(param=hh) -> Iterator[O]:
+    calls = 0
+    def orig(a: int, param=hh) -> Iterator[O]:
         yield O(hh.x)
+        nonlocal calls
+        calls += 1
 
-    list(fun())
-    assert list(fun()) == [O(1)]
+    fun = cachew(tmp_path, depends_on=lambda a, param: (a, param.x))(orig)
+
+    list(fun(123))
+    assert list(fun(123)) == [O(1)]
+    assert calls == 1
 
     # now, change hash. That should cause the composite hash to invalidate and recompute
     hh.x = 2
-    assert list(fun()) == [O(2)]
+    assert list(fun(123)) == [O(2)]
+    assert calls == 2
 
     # should be ok with explicitly passing
-    assert list(fun(param=HackHash(2))) == [O(2)]
+    assert list(fun(123, param=HackHash(2))) == [O(2)]
+    assert calls == 2
+
+    # we don't have to handle the default param in the default hash key
+    fun = cachew(tmp_path)(fun)
+    assert list(fun(456)) == [O(2)]
+    assert calls == 3
+    assert list(fun(456)) == [O(2)]
+    assert calls == 3
+
+    # changing the default should trigger the default (i.e. kwargs) key function to invalidate the cache
+    hh.x = 3
+    assert list(fun(456)) == [O(3)]
+    assert calls == 4
+
+    # you don't have to pass the default parameter explicitly
+    fun = cachew(tmp_path, depends_on=lambda a: a)(orig)
+    assert list(fun(456)) == [O(3)]
+    assert calls == 5
+
+    # but watch out if you forget to handle it!
+    hh.x = 4
+    assert list(fun(456)) == [O(3)]
+    assert calls == 5
 
 
 class U(NamedTuple):
@@ -551,7 +695,7 @@ def fuzz_cachew_impl():
     Insert random sleeps in cachew_impl to increase likelihood of concurrency issues
     """
     import patchy  # type: ignore[import]
-    from .. import cachew_impl
+    from .. import cachew_wrapper
     patch = '''\
 @@ -740,6 +740,11 @@
 
@@ -566,9 +710,9 @@ def fuzz_cachew_impl():
                  logger.debug('hash matched: loading from cache')
                  rows = conn.execute(values_table.select())
 '''
-    patchy.patch(cachew_impl, patch)
+    patchy.patch(cachew_wrapper, patch)
     yield
-    patchy.unpatch(cachew_impl, patch)
+    patchy.unpatch(cachew_wrapper, patch)
 
 
 # TODO fuzz when they start so they enter transaction at different times?
@@ -625,8 +769,9 @@ def test_concurrent_reads(tmp_path: Path, fuzz_cachew_impl):
 
 def test_mcachew(tmp_path: Path):
     # TODO how to test for defensive behaviour?
-    from cachew.misc import mcachew
+    from cachew.extra import mcachew
 
+    # TODO check throw on error
     @mcachew(cache_path=tmp_path / 'cache')
     def func() -> Iterator[str]:
         yield 'one'
@@ -636,12 +781,59 @@ def test_mcachew(tmp_path: Path):
     assert list(func()) == ['one', 'two']
 
 
+def test_defensive(restore_settings):
+    '''
+    Make sure that cachew doesn't crash on misconfiguration
+    '''
+    def orig() -> Iterator[int]:
+        yield 123
+
+    def orig2():
+        yield "x"
+        yield 123
+
+    fun = cachew(bad_arg=123)(orig)
+    assert list(fun()) == [123]
+    assert list(fun()) == [123]
+
+    from cachew.compat import nullcontext
+    for throw in [True, False]:
+        ctx = pytest.raises(Exception) if throw else nullcontext()
+        settings.THROW_ON_ERROR = throw
+
+        with ctx: # type: ignore
+            fun = cachew(cache_path=lambda: 1 + 'bad_path_provider')(orig) # type: ignore
+            assert list(fun()) == [123]
+            assert list(fun()) == [123]
+
+            fun = cachew(cache_path=lambda p: '/tmp/' + str(p))(orig)
+            assert list(fun()) == [123]
+            assert list(fun()) == [123]
+
+            fun = cachew(orig2)
+            assert list(fun()) == ['x', 123]
+            assert list(fun()) == ['x', 123]
+
+            settings.DEFAULT_CACHEW_DIR = '/dev/nonexistent'
+            fun = cachew(orig)
+            assert list(fun()) == [123]
+            assert list(fun()) == [123]
+
+
 def test_recursive(tmp_path: Path):
+    d0 = 0
+    d1 = 1000
     calls = 0
     @cachew(tmp_path)
     def factorials(n: int) -> Iterable[int]:
-        nonlocal calls
+        nonlocal calls, d0, d1
         calls += 1
+
+        if n == 0:
+            d0 = len(inspect.stack(0))
+        if n == 1:
+            d1 = len(inspect.stack(0))
+
         if n == 0:
             yield 1
             return
@@ -655,6 +847,12 @@ def test_recursive(tmp_path: Path):
 
     assert calls == 0
     assert list(factorials(3)) == [1, 1, 2, 6]
+
+    # make sure the recursion isn't eating too much stack
+    # ideally would have 1? not sure if possible without some insane hacking?
+    # todo maybe check stack frame size as well?
+    assert abs(d0 - d1) <= 2
+
     assert calls == 4
     assert list(factorials(3)) == [1, 1, 2, 6]
     assert calls == 4
@@ -711,29 +909,24 @@ def test_recursive_error(tmp_path: Path):
     assert len(list(rec(100))) == 101
 
 
-@pytest.fixture
-def with_exceptions():
-    from cachew.experimental import enable_exceptions, disable_exceptions
-    enable_exceptions()
-    try:
-        yield
-    finally:
-        disable_exceptions()
+def test_exceptions(tmp_path: Path):
+    class X(NamedTuple):
+        a: int
 
-
-def test_exceptions(tmp_path: Path, with_exceptions):
+    d = datetime.strptime('20200102 03:04:05', '%Y%m%d %H:%M:%S')
     @cachew(tmp_path)
     def fun() -> Iterator[Exception]:
-        yield RuntimeError('whatever', 123)
+        yield RuntimeError('whatever', 123, d, X(a=123))
 
     list(fun())
     [e] = fun()
-    assert type(e) == Exception  # sadly we lose type information at the moment..
-    assert e.args == ("('whatever', 123)", )
+    # not sure if there is anything that can be done to preserve type information?
+    assert type(e) == Exception
+    assert e.args == ('whatever', 123, '2020-01-02T03:04:05', 'X(a=123)')
 
 
 # see https://beepb00p.xyz/mypy-error-handling.html#kiss
-def test_result(tmp_path: Path, with_exceptions):
+def test_result(tmp_path: Path):
     @cachew(tmp_path)
     def fun() -> Iterator[Union[Exception, int]]:
         yield 1
@@ -743,9 +936,7 @@ def test_result(tmp_path: Path, with_exceptions):
     [v1, ve, v123] = fun()
     assert v1 == 1
     assert v123 == 123
-    assert ve.args == ("('sad!',)",)
-    # looks bit sad indeed, but whatever...
-    # TODO need to warn perhaps
+    assert ve.args == ('sad!',)
 
 
 def test_version_change(tmp_path: Path):
@@ -823,3 +1014,85 @@ COMMIT;
     # this tests that it doesn't crash
     # for actual version upgrade test see test_version_change
     assert [1, 2, 3] == list(fun())
+
+
+def test_disabled(tmp_path: Path):
+    calls = 0
+    @cachew(tmp_path)
+    def fun() -> Iterator[int]:
+        yield 1
+        yield 2
+        nonlocal calls
+        calls += 1
+
+    assert list(fun()) == [1, 2]
+    assert list(fun()) == [1, 2]
+    assert calls == 1
+
+    from cachew.extra import disabled_cachew
+    with disabled_cachew():
+        assert list(fun()) == [1, 2]
+        assert calls == 2
+        assert list(fun()) == [1, 2]
+        assert calls == 3
+
+
+def test_early_exit(tmp_path: Path):
+    cf = 0
+    @cachew(tmp_path) #  / 'fun', force_file=True)
+    def f() -> Iterator[int]:
+        yield from range(20)
+        nonlocal cf
+        cf += 1
+
+    cg = 0
+    @cachew(tmp_path) # / 'fun', force_file=True)
+    def g() -> Iterator[int]:
+        yield from f()
+        nonlocal cg
+        cg += 1
+
+    assert len(list(islice(g(), 0, 10))) == 10
+    assert cf == 0 # hasn't finished
+    assert cg == 0 # hasn't finished
+
+    # todo not sure if need to check that db is empty?
+   
+    assert len(list(g())) == 20
+    assert cf == 1
+    assert cg == 1
+
+    # should be cached now
+    assert len(list(g())) == 20
+    assert cf == 1
+    assert cg == 1
+
+
+# see https://github.com/sqlalchemy/sqlalchemy/issues/5522#issuecomment-705156746
+def test_early_exit_shutdown(tmp_path: Path):
+    # don't ask... otherwise the exception doesn't appear :shrug:
+    import_hack = '''
+from sqlalchemy import Column
+
+import re
+re.hack = lambda: None
+    '''
+    Path(tmp_path / 'import_hack.py').write_text(import_hack)
+
+    prog = f'''
+import import_hack
+
+import cachew
+cachew.settings.THROW_ON_ERROR = True # todo check with both?
+@cachew.cachew('{tmp_path}', cls=int)
+def fun():
+    yield 0
+
+g = fun()
+e = next(g)
+
+print("FINISHED")
+    '''
+    r = run(['python3', '-c', prog], cwd=tmp_path, stderr=PIPE, stdout=PIPE, check=True)
+    assert r.stdout.strip() == b'FINISHED'
+    assert b'Traceback' not in r.stderr
